package org.artifactory.repo.remote.interceptor;

import org.apache.commons.lang.StringUtils;
import org.apache.http.Header;
import org.apache.http.HttpHeaders;
import org.apache.http.HttpHost;
import org.apache.http.HttpRequest;
import org.apache.http.HttpRequestInterceptor;
import org.apache.http.HttpResponse;
import org.apache.http.HttpResponseInterceptor;
import org.apache.http.client.methods.HttpRequestWrapper;
import org.apache.http.protocol.HttpContext;
import org.artifactory.traffic.OutboundRequestLogger;
import org.artifactory.traffic.entry.OutboundRequestEntry;

import java.util.Optional;

/**
 * Intercepts outbound requests made by Remote Repositories and logs them to artifactory-request-out.log
 *
 * @author Andrey Tashlitsky
 */
public class RemoteRepoRequestResponseInterceptor implements HttpRequestInterceptor, HttpResponseInterceptor {

    private static final String REQUEST_START_TIME = "request_start_time";
    private final String repoName;
    private final String userName;

    public RemoteRepoRequestResponseInterceptor(String repoName, String userName) {
        this.repoName = repoName;
        this.userName = Optional.ofNullable(userName).orElse(StringUtils.EMPTY);
    }

    @Override
    public void process(HttpRequest httpRequest, HttpContext httpContext) {
        httpContext.setAttribute(REQUEST_START_TIME, System.currentTimeMillis());
    }

    @Override
    public void process(HttpResponse httpResponse, HttpContext httpContext) {
        // Get request related info
        long requestStartTime = (long) httpContext.getAttribute(REQUEST_START_TIME);
        HttpRequestWrapper httpRequestWrapper = (HttpRequestWrapper) httpContext.getAttribute("http.request");
        String url = Optional.ofNullable(httpRequestWrapper.getTarget())
                .map(HttpHost::toString)
                .orElse(StringUtils.EMPTY)
                + httpRequestWrapper.getURI().toString();
        long requestContentLength = Optional.ofNullable(httpRequestWrapper.getFirstHeader(HttpHeaders.CONTENT_LENGTH))
                .map(Header::getValue)
                .map(Long::parseLong)
                .orElse(0L);
        // Get response related info
        HttpResponse basicHttpResponse = (HttpResponse) httpContext.getAttribute("http.response");
        int statusCode = basicHttpResponse.getStatusLine().getStatusCode();
        long responseContentLength = Optional.ofNullable(basicHttpResponse.getFirstHeader(HttpHeaders.CONTENT_LENGTH))
                .map(Header::getValue)
                .map(Long::parseLong)
                .orElse(0L);
        // Log the entry
        OutboundRequestEntry outboundRequestEntry = OutboundRequestEntry.builder()
                .repoName(repoName)
                .userName(userName)
                .method(httpRequestWrapper.getMethod())
                .url(url)
                .statusCode(statusCode)
                .requestContentLength(requestContentLength)
                .responseContentLength(responseContentLength)
                .duration(System.currentTimeMillis() - requestStartTime)
                .build();
        OutboundRequestLogger.logEntry(outboundRequestEntry);
    }
}
