/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.layout;

import org.apache.commons.lang3.SerializationUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ArtifactoryRunningMode;
import org.artifactory.addon.webstart.ArtifactWebstartAddon;
import org.artifactory.aol.AolService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.ConstantValues;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.descriptor.repo.ReverseProxyDescriptor;
import org.artifactory.descriptor.repo.releasebundle.ReleaseBundlesRepoDescriptor;
import org.artifactory.request.UrlVerifier;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;
import org.testng.collections.Lists;

import javax.annotation.Nullable;
import java.util.Objects;

import static org.artifactory.addon.AddonType.WEBSTART;
import static org.artifactory.addon.build.BuildAddon.BUILD_INFO_REPO_NAME;
import static org.artifactory.descriptor.repo.SupportBundleRepoDescriptor.SUPPORT_BUNDLE_REPO_NAME;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;
import static org.testng.Assert.fail;

/**
 * @author nadavy
 */
@Test
public class RepositoriesConfigurationChangesInterceptorTest extends ArtifactoryHomeBoundTest {

    private RepositoriesConfigurationChangesInterceptor repoConfigInterceptor;
    private CentralConfigService centralConfigService;
    private AddonsManager addonsManager;
    private AolService aolService;
    private CentralConfigDescriptorImpl desc;
    private InternalArtifactoryContext contextMock;


    @BeforeMethod
    public void setup() {
        addonsManager = mock(AddonsManager.class);
        centralConfigService = mock(CentralConfigService.class);
        aolService = mock(AolService.class);
        contextMock = mock(InternalArtifactoryContext.class);

        when(aolService.isAol()).thenReturn(true);
        when(contextMock.beanForType(AolService.class)).thenReturn(aolService);
        when(contextMock.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        when(addonsManager.getArtifactoryRunningMode()).thenReturn(ArtifactoryRunningMode.PRO);

        desc = new CentralConfigDescriptorImpl();
        desc.addRemoteRepository(addRemoteRepo(true));
        desc.addRemoteRepository(addRemoteRepo(false));
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setKey(BUILD_INFO_REPO_NAME);
        localRepoDescriptor.setType(RepoType.BuildInfo);
        desc.addLocalRepository(localRepoDescriptor);
        when(centralConfigService.getDescriptor()).thenReturn(desc);
        repoConfigInterceptor = new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                createUrlVerifier());

        ArtifactoryContextThreadBinder.bind(contextMock);
    }

    @AfterMethod
    public void clenaUp() {
        ArtifactoryContextThreadBinder.unbind();
    }

    public void testHandleReverseProxySettingsOssRunningMode() {
        when(addonsManager.getArtifactoryRunningMode()).thenReturn(ArtifactoryRunningMode.OSS);
        RepositoriesConfigurationChangesInterceptor repoConfChangesInterceptor =
                new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                        createUrlVerifier());
        CentralConfigDescriptorImpl newDescriptor = new CentralConfigDescriptorImpl();
        newDescriptor.addReverseProxy(new ReverseProxyDescriptor());
        repoConfChangesInterceptor.handleReverseProxySettings(newDescriptor);
        Assert.assertEquals(newDescriptor.getReverseProxies().size(), 0);
    }

    public void testHandleReverseProxySettingsProRunningMode() {
        when(addonsManager.getArtifactoryRunningMode()).thenReturn(ArtifactoryRunningMode.PRO);
        RepositoriesConfigurationChangesInterceptor repoConfChangesInterceptor =
                new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                        createUrlVerifier());
        CentralConfigDescriptorImpl newDescriptor = new CentralConfigDescriptorImpl();
        newDescriptor.addReverseProxy(new ReverseProxyDescriptor());
        repoConfChangesInterceptor.handleReverseProxySettings(newDescriptor);
        Assert.assertEquals(newDescriptor.getReverseProxies().size(), 1);
    }

    public void testHandleReverseProxySettingsJcrRunningMode() {
        when(addonsManager.getArtifactoryRunningMode()).thenReturn(ArtifactoryRunningMode.JCR);
        RepositoriesConfigurationChangesInterceptor repoConfChangesInterceptor =
                new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                        createUrlVerifier());
        CentralConfigDescriptorImpl newDescriptor = new CentralConfigDescriptorImpl();
        newDescriptor.addReverseProxy(new ReverseProxyDescriptor());
        repoConfChangesInterceptor.handleReverseProxySettings(newDescriptor);
        Assert.assertEquals(newDescriptor.getReverseProxies().size(), 1);
    }

    public void testReloadWithAddons() {
        enableAddons(true, true, true);
        repoConfigInterceptor.onBeforeSave(desc);
        assertEquals(countValidCertificates(), 1);
    }

    public void testReloadWithoutAddons() {
        enableAddons(false, true, true);
        repoConfigInterceptor.onBeforeSave(desc);
        assertEquals(countValidCertificates(), 2);
    }

    public void testBlacklistedUrls() {
        enableAddons(true, true, true);
        getBound().setProperty(ConstantValues.whitelistRemoteRepoUrls, "http://localhost");
        repoConfigInterceptor = new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                createUrlVerifier());
        CentralConfigDescriptorImpl clone = SerializationUtils.clone(desc);
        clone.addRemoteRepository(addRemoteRepo("http://localhost:80/artifactory"));
        repoConfigInterceptor.onBeforeSave(clone);
        boolean blackedOut =
                clone.getRemoteRepositoriesMap().get("Remote-http://localhost:80/artifactory").isBlackedOut();
        assertFalse(blackedOut);
        // save will be ok

        getBound().setProperty(ConstantValues.whitelistRemoteRepoUrls, null);
        repoConfigInterceptor = new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                createUrlVerifier());
        desc.addRemoteRepository(addRemoteRepo("http://localhost:8080/artifactory"));
        clone = SerializationUtils.clone(desc);
        clone.addRemoteRepository(addRemoteRepo("http://google.com/artifactory"));
        repoConfigInterceptor.onBeforeSave(clone);
        blackedOut = clone.getRemoteRepositoriesMap().get("Remote-http://localhost:8080/artifactory").isBlackedOut();
        assertTrue(blackedOut);
        // save will be ok - but repository should be blacked out

        clone = SerializationUtils.clone(desc);
        clone.addRemoteRepository(addRemoteRepo("http://localhost/artifactory"));
        try {
            repoConfigInterceptor.onBeforeSave(clone);
        } catch (IllegalArgumentException expected) {
            // this is good!
            assertEquals(expected.getMessage(), "Found a remote repository(ies) containing blacklisted URLs");
            return;
        }
        fail("Should have found a remote repository(ies) containing blacklisted URLs");
    }

    public void testNotAol() {
        try {
            enableAddons(true, true, true);
            when(aolService.isAol()).thenReturn(false);
            repoConfigInterceptor = new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                    new UrlVerifier());
            CentralConfigDescriptorImpl clone = SerializationUtils.clone(desc);
            clone.getRemoteRepositoriesMap().get("Repo-Valid-Cert-true").setUrl("http://localhost:8080/artifactory");
            repoConfigInterceptor.onBeforeSave(clone);
        } finally {
            when(aolService.isAol()).thenReturn(true);
        }
    }

    public void testBlockedUrlsChangeUrl() {
        enableAddons(true, true, true);
        repoConfigInterceptor = new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                createUrlVerifier());
        CentralConfigDescriptorImpl clone = SerializationUtils.clone(desc);
        clone.getRemoteRepositoriesMap().get("Repo-Valid-Cert-true").setUrl("http://google.com/artifactory");
        repoConfigInterceptor.onBeforeSave(clone);

        enableAddons(true, true, true);
        repoConfigInterceptor = new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                createUrlVerifier());
        clone = SerializationUtils.clone(desc);
        clone.getRemoteRepositoriesMap().get("Repo-Valid-Cert-true").setUrl("http://localhost:8080/artifactory");
        try {
            repoConfigInterceptor.onBeforeSave(clone);
        } catch (IllegalArgumentException expected) {
            // this is good!
            assertEquals(expected.getMessage(), "Found a remote repository(ies) containing blacklisted URLs");
            return;
        }
        fail("Update URL should have been blocked!");
    }

    public void testEnableCdnRedirectForEnterprisePlusAol() {
        enableAddons(true, true, true);

        RepositoriesConfigurationChangesInterceptor repoConfChangesInterceptor =
                new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                        createUrlVerifier());
        LocalRepoDescriptor repoDescriptor = createLocalDescriptorRedirectEnabled(null);
        desc.addLocalRepository(repoDescriptor);

        repoConfChangesInterceptor.onBeforeSave(desc);
        // we expect download redirect to be off but cdn redirect still on
        LocalRepoDescriptor resultRepoDescriptor = desc.getLocalRepositoriesMap().get("cdn-local-repo");
        assertTrue(resultRepoDescriptor.isDownloadRedirect());
        assertTrue(resultRepoDescriptor.isCdnRedirect());
    }

    public void testEnableCdnRedirectForEnterpriseAol() {
        enableAddons(true, false, true);

        RepositoriesConfigurationChangesInterceptor repoConfChangesInterceptor =
                new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                        createUrlVerifier());
        LocalRepoDescriptor repoDescriptor = createLocalDescriptorRedirectEnabled(null);
        desc.addLocalRepository(repoDescriptor);
        when(centralConfigService.isContextCreated()).thenReturn(true);
        repoConfChangesInterceptor.onBeforeSave(desc);
        // we expect download redirect to be off but cdn redirect still on
        LocalRepoDescriptor resultRepoDescriptor = desc.getLocalRepositoriesMap().get("cdn-local-repo");
        assertNull(resultRepoDescriptor.getDownloadRedirectConfig());
        assertTrue(resultRepoDescriptor.isCdnRedirect());
    }

    public void testEnableCdnRedirectForProAol() {
        enableAddons(true, false, false);

        RepositoriesConfigurationChangesInterceptor repoConfChangesInterceptor =
                new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                        createUrlVerifier());
        LocalRepoDescriptor repoDescriptor = createLocalDescriptorRedirectEnabled(null);
        desc.addLocalRepository(repoDescriptor);
        when(centralConfigService.isContextCreated()).thenReturn(true);
        repoConfChangesInterceptor.onBeforeSave(desc);

        LocalRepoDescriptor resultRepoDescriptor = desc.getLocalRepositoriesMap().get("cdn-local-repo");
        assertNull(resultRepoDescriptor.getDownloadRedirectConfig());
        assertTrue(resultRepoDescriptor.isCdnRedirect());
    }

    public void testDisableCdnRedirectOnPrem() {
        when(aolService.isAol()).thenReturn(false);
        enableAddons(true, true, false);

        RepositoriesConfigurationChangesInterceptor repoConfChangesInterceptor =
                new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                        createUrlVerifier());
        LocalRepoDescriptor repoDescriptor = createLocalDescriptorRedirectEnabled(null);
        desc.addLocalRepository(repoDescriptor);

        repoConfChangesInterceptor.onBeforeSave(desc);

        LocalRepoDescriptor resultRepoDescriptor = desc.getLocalRepositoriesMap().get("cdn-local-repo");
        assertNull(resultRepoDescriptor.getCdnRedirectRepoConfig());
        assertFalse(resultRepoDescriptor.isCdnRedirect());
    }

    @Test(expectedExceptions = IllegalArgumentException.class)
    public void testStrictPolicy() {
        // strict policy will fail by default, in cases of unresolvable URLs
        enableAddons(true, true, true);
        getBound().setProperty(ConstantValues.remoteRepoBlockUrlStrictPolicy, Boolean.toString(true));
        repoConfigInterceptor = new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                createUrlVerifier());
        CentralConfigDescriptorImpl clone = SerializationUtils.clone(desc);
        clone.addRemoteRepository(addRemoteRepo("\\o//"));
        repoConfigInterceptor.onBeforeSave(clone);
    }

    public void testForbiddenSupportBundleRepo() {
        // strict policy will fail by default, in cases of unresolvable URLs
        enableAddons(true, true, true);
        repoConfigInterceptor = new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                createUrlVerifier());
        CentralConfigDescriptorImpl clone = SerializationUtils.clone(desc);

        clone.addLocalRepository(createLocalDescriptorRedirectEnabled(null));
        assertFalse(repoConfigInterceptor.isForbiddenSupportBundleRepoExist(clone));
        clone.addRemoteRepository(createRemoteRepo("test-remote"));
        assertFalse(repoConfigInterceptor.isForbiddenSupportBundleRepoExist(clone));
        clone.addFederatedRepository(createFederatedRepo("test-fed"));
        assertFalse(repoConfigInterceptor.isForbiddenSupportBundleRepoExist(clone));
        // add RepoType.Support local type (should be allowed)
        LocalRepoDescriptor localSupportBundleDesc =
                createLocalDescriptorRedirectEnabled(SUPPORT_BUNDLE_REPO_NAME);
        localSupportBundleDesc.setType(RepoType.Support);
        clone.addLocalRepository(localSupportBundleDesc);
        assertFalse(repoConfigInterceptor.isForbiddenSupportBundleRepoExist(clone));
        // add non RepoType.Support local type (should fail)
        localSupportBundleDesc.setType(RepoType.Generic);
        clone.addLocalRepository(localSupportBundleDesc);
        assertTrue(repoConfigInterceptor.isForbiddenSupportBundleRepoExist(clone));
    }

    public void testForbiddenDistributionRepo() {
        // strict policy will fail by default, in cases of unresolvable URLs
        enableAddons(true, true, true);
        repoConfigInterceptor = new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                createUrlVerifier());
        CentralConfigDescriptorImpl clone = SerializationUtils.clone(desc);

        clone.addLocalRepository(createLocalDescriptorRedirectEnabled(null));
        assertFalse(repoConfigInterceptor.isForbiddenReleaseBundlesRepoExist(clone));
        clone.addRemoteRepository(createRemoteRepo("test-remote"));
        assertFalse(repoConfigInterceptor.isForbiddenReleaseBundlesRepoExist(clone));
        clone.addFederatedRepository(createFederatedRepo("test-fed"));
        assertFalse(repoConfigInterceptor.isForbiddenReleaseBundlesRepoExist(clone));
        clone.addLocalRepository(createLocalDescriptorRedirectEnabled(SUPPORT_BUNDLE_REPO_NAME));
        assertFalse(repoConfigInterceptor.isForbiddenReleaseBundlesRepoExist(clone));
    }

    public void testLoadRedirectDownloadWithNoContextCreated() {
        enableAddons(true, true, false);
        RepositoriesConfigurationChangesInterceptor repoConfChangesInterceptor =
                new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                        createUrlVerifier());
        LocalRepoDescriptor repoDescriptor = createLocalDescriptorRedirectEnabled(null);
        desc.addLocalRepository(repoDescriptor);
        when(addonsManager.isClusterEnterprisePlus()).thenReturn(false);
        when(addonsManager.isEdgeLicensed()).thenReturn(true);
        when(centralConfigService.isContextCreated()).thenReturn(false);
        repoConfChangesInterceptor.onBeforeSave(desc);

        LocalRepoDescriptor resultRepoDescriptor = desc.getLocalRepositoriesMap().get("cdn-local-repo");
        assertTrue(resultRepoDescriptor.getDownloadRedirectConfig().isEnabled(),
                "Should not turn off redirect download option if not all beans were loaded and configuration service" +
                        "context wasn't created ");
    }

    public void testLoadRedirectDownloadWithContextCreatedWithEdgeLicensed() {
        enableAddons(true, true, false);
        RepositoriesConfigurationChangesInterceptor repoConfChangesInterceptor =
                new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                        createUrlVerifier());
        LocalRepoDescriptor repoDescriptor = createLocalDescriptorRedirectEnabled(null);
        desc.addLocalRepository(repoDescriptor);
        when(addonsManager.isClusterEnterprisePlus()).thenReturn(false);
        when(addonsManager.isEdgeLicensed()).thenReturn(true);
        when(centralConfigService.isContextCreated()).thenReturn(true);
        repoConfChangesInterceptor.onBeforeSave(desc);

        LocalRepoDescriptor resultRepoDescriptor = desc.getLocalRepositoriesMap().get("cdn-local-repo");
        assertTrue(resultRepoDescriptor.getDownloadRedirectConfig().isEnabled(),
                "Should not turn off redirect download option with edge licensed");
    }

    public void testLoadRedirectDownloadWithContextCreatedWithEnterprisePlus() {
        enableAddons(true, true, false);
        RepositoriesConfigurationChangesInterceptor repoConfChangesInterceptor =
                new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                        createUrlVerifier());
        LocalRepoDescriptor repoDescriptor = createLocalDescriptorRedirectEnabled(null);
        desc.addLocalRepository(repoDescriptor);
        when(addonsManager.isClusterEnterprisePlus()).thenReturn(true);
        when(addonsManager.isEdgeLicensed()).thenReturn(false);
        when(centralConfigService.isContextCreated()).thenReturn(true);
        repoConfChangesInterceptor.onBeforeSave(desc);

        LocalRepoDescriptor resultRepoDescriptor = desc.getLocalRepositoriesMap().get("cdn-local-repo");
        assertTrue(resultRepoDescriptor.getDownloadRedirectConfig().isEnabled(),
                "Should not turn off redirect download option with enterprise plus");
    }


    public void testLoadRedirectDownloadWithOutEnterprisePlusOrEdgeLicense() {
        enableAddons(true, true, false);
        RepositoriesConfigurationChangesInterceptor repoConfChangesInterceptor =
                new RepositoriesConfigurationChangesInterceptor(addonsManager, centralConfigService,
                        createUrlVerifier());
        LocalRepoDescriptor repoDescriptor = createLocalDescriptorRedirectEnabled(null);
        desc.addLocalRepository(repoDescriptor);
        when(addonsManager.isClusterEnterprisePlus()).thenReturn(false);
        when(addonsManager.isEdgeLicensed()).thenReturn(false);
        when(centralConfigService.isContextCreated()).thenReturn(true);
        repoConfChangesInterceptor.onBeforeSave(desc);
        LocalRepoDescriptor resultRepoDescriptor = desc.getLocalRepositoriesMap().get("cdn-local-repo");
        assertNull(resultRepoDescriptor.getDownloadRedirectConfig(), "Should turn off redirect download option");
    }

    /**
     * Count the number of remote repositories with a client TLS certificates
     */
    private long countValidCertificates() {
        return desc.getRemoteRepositoriesMap().values().stream()
                .filter(r -> r instanceof HttpRepoDescriptor)
                .map(r -> (HttpRepoDescriptor) r)
                .map(HttpRepoDescriptor::getClientTlsCertificate)
                .filter(Objects::nonNull)
                .count();
    }

    /**
     * Add remote repository with a tls certificate. the certificate may or may not be in the keystore
     */
    private HttpRepoDescriptor addRemoteRepo(boolean withValidCert) {
        HttpRepoDescriptor remoteRepoDescriptor = new HttpRepoDescriptor();
        remoteRepoDescriptor.setKey("Repo-Valid-Cert-" + withValidCert);
        remoteRepoDescriptor.setUrl("http://someurl.com");
        remoteRepoDescriptor.setClientTlsCertificate(withValidCert ? "ValidTls" : "NotValidTLS");
        return remoteRepoDescriptor;
    }

    /**
     * Add remote repository with a URL
     */
    private HttpRepoDescriptor addRemoteRepo(String url) {
        HttpRepoDescriptor remoteRepoDescriptor = new HttpRepoDescriptor();
        remoteRepoDescriptor.setKey("Remote-" + url);
        remoteRepoDescriptor.setUrl(url);
        return remoteRepoDescriptor;
    }

    private LocalRepoDescriptor createLocalDescriptorRedirectEnabled(@Nullable String repoKey) {
        LocalRepoDescriptor repoDescriptor = new LocalRepoDescriptor();
        repoDescriptor.setCdnRedirect(true);
        repoDescriptor.setDownloadRedirect(true);
        if (repoKey == null) {
            repoDescriptor.setKey("cdn-local-repo");
        } else {
            repoDescriptor.setKey(repoKey);
        }
        return repoDescriptor;
    }

    /**
     * Enabled or disable the addons. if enabled returns a list with 1 valid certificate in the keystore, return null
     * otherwise (like in CoreAddonsImpl)
     */
    private void enableAddons(boolean enabled, boolean isEntPlus, boolean isEnt) {
        when(addonsManager.isEdgeLicensed()).thenReturn(false);
        when(addonsManager.isClusterEnterprisePlus()).thenReturn(isEntPlus);
        when(addonsManager.isEnterpriseLicensed()).thenReturn(isEnt);
        when(addonsManager.isAddonSupported(WEBSTART)).thenReturn(enabled);
        ArtifactWebstartAddon webstartAddon = mock(ArtifactWebstartAddon.class);
        when(addonsManager.addonByType(ArtifactWebstartAddon.class)).thenReturn(webstartAddon);
        when(webstartAddon.getSslCertNames()).thenReturn(enabled ? Lists.newArrayList("ValidTls") : null);
    }

    private FederatedRepoDescriptor createFederatedRepo(String repoKey) {
        FederatedRepoDescriptor repoDescriptor = new FederatedRepoDescriptor();
        repoDescriptor.setKey(repoKey);
        repoDescriptor.setType(RepoType.Generic);
        return repoDescriptor;
    }

    private RemoteRepoDescriptor createRemoteRepo(String repoKey) {
        HttpRepoDescriptor repoDescriptor = new HttpRepoDescriptor();
        repoDescriptor.setKey(repoKey);
        repoDescriptor.setType(RepoType.Generic);
        return repoDescriptor;
    }

    private ReleaseBundlesRepoDescriptor createReleaseBundleRepoDesc(String repoKey) {
        ReleaseBundlesRepoDescriptor repoDescriptor = new ReleaseBundlesRepoDescriptor();
        repoDescriptor.setKey(repoKey);
        repoDescriptor.setType(RepoType.Generic);
        return repoDescriptor;
    }

    private UrlVerifier createUrlVerifier() {
        return new UrlVerifier();
    }
}