/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.security.props.auth;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import lombok.AllArgsConstructor;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.ToString;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.crypto.CryptoHelper;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.security.props.auth.model.TokenKeyValue;
import org.jfrog.security.crypto.JFrogCryptoHelper;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.annotation.Nonnull;
import java.security.GeneralSecurityException;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import static com.google.common.collect.Sets.newHashSet;

/**
 * @author Chen Keinan
 */
@Component
public abstract class PropsTokenManager implements TokenManager, EncryptedTokenManager {
    static final Logger log = LoggerFactory.getLogger(PropsTokenManager.class);
    private final Set<String> propKeys = newHashSet();
    private final LoadingCache<TokensCacheKey, String> tokensCache;

    @Autowired
    private UserGroupService userGroupService;

    PropsTokenManager() {
        propKeys.add(getPropKey());
        tokensCache = CacheBuilder.newBuilder()
                .expireAfterWrite(ConstantValues.propsTokenCacheExpirySeconds.getInt(), TimeUnit.SECONDS)
                .maximumSize(ConstantValues.propsTokenCacheSize.getInt())
                .build(new CacheLoader<TokensCacheKey, String>() {
                    @Override
                    public String load(@Nonnull TokensCacheKey key) {
                        return userGroupService.getPropsToken(key.getUsername(), key.getPropKey());
                    }
                });
    }

    /**
     * return props key for each token type (oauth , apiKey and etc)
     *
     * @return key
     */
    public abstract String getPropKey();

    @Override
    public TokenKeyValue createToken(String userName) {
        TokenKeyValue token = null;
        String tokenValue = null;
        String key = getPropKey();
        try {
            tokenValue = JFrogCryptoHelper.generateUniqueApiKeyToken();
            boolean tokenPropCreated = userGroupService.createPropsToken(userName, key, tokenValue);
            if (tokenPropCreated) {
                token = new TokenKeyValue(key, tokenValue);
            }
        } catch (GeneralSecurityException e) {
            log.debug("error with generating token for user: '{}' with key {}", userName, key, e);
        } catch (Exception e) {
            log.debug("error with adding token for user: '{}' with key {} and value {}", userName, key, tokenValue, e);
        }
        return token;
    }

    @Override
    public TokenKeyValue addExternalToken(String userName, String tokenValue) {
        TokenKeyValue token = null;
        String key = getPropKey();
        try {
            boolean propsToken = userGroupService.createPropsToken(userName, key, tokenValue);
            if (propsToken) {
                token = new TokenKeyValue(key, tokenValue);
            }
        } catch (Exception e) {
            log.debug("error with adding external token for user: '{}' with key {} and value {}", userName, key,
                    tokenValue, e);
        }
        return token;
    }

    @Override
    public TokenKeyValue refreshToken(String userName) {
        TokenKeyValue token = null;
        try {
            String value = JFrogCryptoHelper.generateUniqueApiKeyToken();
            token = updateToken(userName, value);
        } catch (GeneralSecurityException e) {
            log.debug("error with refreshing token for user: '{}'", userName, e);
        }
        return token;
    }

    @Override
    public TokenKeyValue updateToken(String userName, String value) {
        String key = getPropKey();
        TokenKeyValue token = null;
        try {
            boolean propsToken = userGroupService.updatePropsToken(userName, key, value);
            if (propsToken) {
                invalidateTokensCache(userName);
                token = new TokenKeyValue(key, value);
            }
        } catch (Exception e) {
            log.debug("error with updating token for user: '{}' with key {} and value {}", userName, key, value, e);
        }
        return token;
    }

    @Override
    public TokenKeyValue getToken(String userName, boolean useCache) {
        String key = getPropKey();
        TokenKeyValue token = null;
        try {
            String value = useCache ?
                    tokensCache.get(new TokensCacheKey(userName, key)) :
                    userGroupService.getPropsToken(userName, key);
            if (value != null) {
                token = new TokenKeyValue(key, CryptoHelper.decryptIfNeeded(ArtifactoryHome.get(), value));
            }
        } catch (CacheLoader.InvalidCacheLoadException | ExecutionException e) {
            // Can happen if using cache and value == null or if an error has occurred
            log.debug("Could not get props token for user '{}', key {}", userName, key, e);
        }
        return token;
    }

    @Override
    public boolean revokeToken(String userName) {
        boolean tokenRevokeSucceeded = false;
        String key = getPropKey();
        try {
            invalidateTokensCache(userName);
            tokenRevokeSucceeded = userGroupService.revokePropsToken(userName, key, shouldNotifyListeners());
        } catch (Exception e) {
            log.debug("error with revoking token for user: '{}' with key {}", userName, key, e);
        }
        return tokenRevokeSucceeded;
    }

    @Override
    public boolean revokeAllTokens() {
        boolean tokenRevokeSucceeded = false;
        String key = getPropKey();
        try {
            tokensCache.invalidateAll(tokensCache.asMap().keySet().stream().filter(tokensCacheKey -> tokensCacheKey
                    .getPropKey().equals(getPropKey())).collect(Collectors.toSet()));
            userGroupService.revokeAllPropsTokens(key, shouldNotifyListeners());
            tokenRevokeSucceeded = true;
        } catch (Exception e) {
            log.debug("error with revoking all tokens with key {}", key, e);
        }
        return tokenRevokeSucceeded;
    }

    @Override
    public Set<String> getPropKeys() {
        return propKeys;
    }

    /**
     * Invalidates the token for a username. This is called also upon updateToken and revokeToken
     */
    private void invalidateTokensCache(String username) {
        tokensCache.invalidate(new TokensCacheKey(username, getPropKey()));
    }

    /**
     * Returns true if the security listeners should be notified upon revoking tokens
     */
    protected boolean shouldNotifyListeners() {
        return true;
    }

    @AllArgsConstructor
    @Getter
    @EqualsAndHashCode
    @ToString
    private static class TokensCacheKey {
        private final String username;
        private final String propKey;
    }
}
