package org.artifactory.storage.db;

import org.artifactory.storage.db.version.ArtifactoryDBVersion;
import org.artifactory.storage.db.version.converter.DBConverter;
import org.artifactory.storage.db.version.converter.LockingDBSqlConverter;
import org.artifactory.util.CollectionUtils;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import java.util.Arrays;
import java.util.List;

import static org.mockito.Mockito.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.mock;

/**
 * Verify that DatabaseConverterAdapter and  LockingDBConverterAdapter process the
 * right DBConverters
 * <p>DatabaseConverterAdapter: all</p>
 * <p>LockingDBConverterAdapter: only converters that supports locking db</p>
 */

public class DatabaseConverterAdapterTest {

    private DatabaseConverterAdapter databaseConverterAdapter = null;
    private LockingDBConverterAdapter lockingDBConverterAdapter = null;


    @Before
    public void setUp() {
        databaseConverterAdapter = mock(DatabaseConverterAdapter.class);
        doAnswer(invocation -> {
            return invocation.callRealMethod();
        }).when(databaseConverterAdapter).getDBConverters(any(ArtifactoryDBVersion.class));

        lockingDBConverterAdapter = mock(LockingDBConverterAdapter.class);
        doAnswer(invocation -> {
            return invocation.callRealMethod();
        }).when(lockingDBConverterAdapter).getDBConverters(any(ArtifactoryDBVersion.class));

    }

    /**
     * This test verify that DatabaseConverterAdapter.getDBConverters(...) returns all types of converters
     * <p>Converters which not support locking db<p/>
     * <p>Converters which support locking db<p/>
     *
     * @throws Exception
     */
    @Test
    public void testDatabaseConverterAdapterGetDBConvertersReturnsAllDBConverters() throws Exception {
        List<ArtifactoryDBVersion> versions = Arrays.asList(ArtifactoryDBVersion.v224, ArtifactoryDBVersion.v225);
        for (ArtifactoryDBVersion dbVersion : versions) {
            DBConverter[] converters = databaseConverterAdapter.getDBConverters(dbVersion);
            Assert.assertFalse(CollectionUtils.isNullOrEmpty(converters));
        }
    }

    /**
     * This test verify that LockingDBConverterAdapter.getDBConverters(...) returns only Converters
     * Converters which support locking db
     *
     * @throws Exception
     */
    @Test
    public void testLockingDBConverterAdapterGetDBConvertersReturnsOnlySupportingLockingDBConverters()
            throws Exception {
        List<ArtifactoryDBVersion> versions = Arrays.asList(ArtifactoryDBVersion.v224, ArtifactoryDBVersion.v225);
        for (ArtifactoryDBVersion dbVersion : versions) {
            DBConverter[] converters = lockingDBConverterAdapter.getDBConverters(dbVersion);
            if (!CollectionUtils.isNullOrEmpty(converters)) {
                for (DBConverter converter : converters) {
                    Assert.assertTrue(converter.getClass().getSimpleName() + " not supports locking db",
                            converter instanceof LockingDBSqlConverter);
                }
            }
        }
    }
}
