package org.artifactory.ui.rest.service.artifacts.browse.generic;

import org.apache.http.HttpStatus;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.CoreAddons;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.BaseBrowsableItem;
import org.artifactory.api.repo.BrowsableItem;
import org.artifactory.api.repo.BrowsableItemCriteria;
import org.artifactory.api.repo.RepositoryBrowsingService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.repo.RootNodesFilterResult;
import org.artifactory.api.repo.exception.FolderExpectedException;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.ui.rest.model.artifacts.browse.generic.NativeBrowserChild;
import org.artifactory.ui.rest.model.artifacts.browse.generic.NativeBrowserModel;
import org.artifactory.ui.rest.resource.artifacts.browse.generic.NativeBrowserResource;
import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentMatchers;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;

import javax.servlet.http.HttpServletRequest;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import static org.artifactory.common.ConstantValues.uiHideChecksums;

public class BrowseNativeModelServiceTest extends ArtifactoryHomeBoundTest {

    @Mock
    private RepositoryService repositoryService;
    @Mock
    private RepositoryBrowsingService repoBrowsingService;
    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private CentralConfigService centralConfigService;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private CoreAddons coreAddons;
    @Mock
    private ArtifactoryContext artifactoryContext;
    @InjectMocks
    private BrowseNativeModelService browseNativeService;

    @Before
    public void setUp() {
        MockitoAnnotations.openMocks(this);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        Mockito.when(artifactoryContext.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        Mockito.when(addonsManager.addonByType(CoreAddons.class)).thenReturn(coreAddons);
        Mockito.when(coreAddons.getListBrowsingVersion()).thenReturn("1.0.0");
        bindArtifactoryHome();
        homeStub.setProperty(uiHideChecksums, "FALSE");
    }

    @After
    public void tearDown() {
        ArtifactoryHome.unbind();
    }

    @Test
    public void testWithEmptyRepoPath() {
        ArtifactoryRestRequest restRequest = Mockito.mock(ArtifactoryRestRequest.class);
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        browseNativeService.execute(restRequest, response);

        Assert.assertEquals(HttpStatus.SC_NOT_FOUND, response.getResponseCode());
    }

    @Test
    public void testWithUnauthorizedUser() {
        Mockito.when(authorizationService.canRead(ArgumentMatchers.any(RepoPath.class))).thenReturn(false);
        Mockito.when(repositoryService.getAllRepoKeys()).thenReturn(Set.of("generic"));
        ArtifactoryRestRequest restRequest = Mockito.mock(ArtifactoryRestRequest.class);
        Mockito.when(restRequest.getPathParamByKey(NativeBrowserResource.REPO_PATH)).thenReturn("generic/somePath");
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        browseNativeService.execute(restRequest, response);

        Assert.assertEquals(HttpStatus.SC_NOT_FOUND, response.getResponseCode());
    }

    @Test
    public void testWithUnExistRepoKey() {
        Mockito.when(authorizationService.canRead(ArgumentMatchers.any(RepoPath.class))).thenReturn(true);
        Mockito.when(repositoryService.getAllRepoKeys()).thenReturn(Set.of(""));
        ArtifactoryRestRequest restRequest = Mockito.mock(ArtifactoryRestRequest.class);
        Mockito.when(restRequest.getPathParamByKey(NativeBrowserResource.REPO_PATH)).thenReturn("generic/somePath");
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        browseNativeService.execute(restRequest, response);

        Assert.assertEquals(HttpStatus.SC_NOT_FOUND, response.getResponseCode());
    }

    @Test
    public void testWithFileDownload() {
        RepoPath repoPath = RepoPathFactory.create("generic/folder/file");
        Mockito.when(repositoryService.getAllRepoKeys()).thenReturn(Set.of(repoPath.getRepoKey()));
        Mockito.when(authorizationService.canRead(ArgumentMatchers.any(RepoPath.class))).thenReturn(true);
        ArtifactoryRestRequest restRequest = Mockito.mock(ArtifactoryRestRequest.class);
        Mockito.when(restRequest.getServletRequest()).thenReturn(Mockito.mock(HttpServletRequest.class));
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        Mockito.when(restRequest.getPathParamByKey(NativeBrowserResource.REPO_PATH)).thenReturn(repoPath.toPath());
        Mockito.when(repositoryService.localFederatedCachedRepoDescriptorByKey(repoPath.getRepoKey()))
                .thenReturn(new LocalRepoDescriptor());
        Mockito.when(repoBrowsingService
                .getLocalRepoBrowsableChildren(Mockito.any(BrowsableItemCriteria.class), Mockito.anyBoolean(),
                        Mockito.any(RootNodesFilterResult.class))).thenThrow(new FolderExpectedException(repoPath));

        browseNativeService.execute(restRequest, response);
        NativeBrowserModel model = (NativeBrowserModel) response.getIModel();
        Assert.assertEquals("generic", model.getRepo());
        Assert.assertEquals("folder/file", model.getPath());
        Assert.assertFalse(model.isRoot());
        Assert.assertFalse(model.isFolder());
        Assert.assertNull(model.getChildren());
        Assert.assertNull(model.getFooter());
    }

    @Test
    public void testWithRealRepoPath() {
        Mockito.when(repositoryService.getAllRepoKeys()).thenReturn(Set.of("generic"));
        Mockito.when(authorizationService.canRead(ArgumentMatchers.any(RepoPath.class))).thenReturn(true);
        ArtifactoryRestRequest restRequest = Mockito.mock(ArtifactoryRestRequest.class);
        Mockito.when(restRequest.getServletRequest()).thenReturn(Mockito.mock(HttpServletRequest.class));
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        List<BaseBrowsableItem> expectedList = new LinkedList<>();
        expectedList
                .add(new BrowsableItem("folder", true, 123L, 1234L, 200L, RepoPathFactory.create("generic", "folder")));
        expectedList
                .add(new BrowsableItem("file", false, 124L, 1235L, 30L, RepoPathFactory.create("generic", "file")));
        Mockito.when(restRequest.getPathParamByKey(NativeBrowserResource.REPO_PATH)).thenReturn("generic/somePath");
        Mockito.when(repositoryService.localFederatedCachedRepoDescriptorByKey("generic"))
                .thenReturn(new LocalRepoDescriptor());
        Mockito.when(repoBrowsingService
                .getLocalRepoBrowsableChildren(Mockito.any(BrowsableItemCriteria.class), Mockito.anyBoolean(),
                        Mockito.any(RootNodesFilterResult.class))).thenReturn(expectedList);
        browseNativeService.execute(restRequest, response);

        Assert.assertEquals(HttpStatus.SC_OK, response.getResponseCode());
        NativeBrowserModel model = (NativeBrowserModel) response.getIModel();
        Assert.assertEquals("generic", model.getRepo());
        Assert.assertEquals("somePath", model.getPath());
        Assert.assertEquals("1.0.0 Server at null Port 0", model.getFooter());
        Assert.assertFalse(model.isRoot());
        Assert.assertTrue(model.isFolder());
        List<NativeBrowserChild> items = model.getChildren();
        Assert.assertEquals("folder", items.get(0).getName());
        Assert.assertEquals(200L, items.get(0).getSize());
        Assert.assertEquals("file", items.get(1).getName());
        Assert.assertEquals(30L, items.get(1).getSize());
    }
}